#
# This file is part of LiteDRAM.
#
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from math import ceil
from operator import and_

from migen import *
from migen.genlib.misc import WaitTimer

from litex.soc.interconnect.csr import AutoCSR, CSR, CSRStatus, CSRStorage

from litedram.common import *
from litedram.phy.utils import chunks, bitpattern
from litedram.phy.dfi import Interface as DFIInterface
from litedram.phy.rpc.commands import DFIAdapter


class ShiftRegister(Module):
    def __init__(self, n, i=None):
        if i is None:
            i = Signal()
        assert len(i) == 1

        self.i = i
        self.sr = sr = Signal(n)
        last = Signal.like(sr)

        self.comb += sr.eq(Cat(i, last))
        self.sync += last.eq(sr)

    def __getitem__(self, key):
        return self.sr[key]


class RPCPads:
    _layout = [
        ("clk_p",  1),
        ("clk_n",  1),
        ("cs_n",   1),
        ("dqs_p",  1),  # may be 2 (hardware option; 2-bit DQS strobes DB by bytes: [0:7], [8:15])
        ("dqs_n",  1),  # may be 2
        ("stb",    1),
        ("db",    16),
    ]

    def __init__(self, pads):
        self.map(pads)
        for pad, width in self._layout:
            assert len(getattr(self, pad)) >= width, \
                "Pad {} has width {} < {}".format(pad, len(getattr(self, pad)), width)

    # reimplement if a specific mapping is needed
    def map(self, pads):
        for pad, _ in self._layout:
            setattr(self, pad, getattr(pads, pad))


class BasePHY(Module, AutoCSR):
    def __init__(self, pads, sys_clk_freq, write_ser_latency, read_des_latency, phytype):
        # TODO: pads groups for multiple chips
        #  pads = PHYPadsCombiner(pads)
        if not isinstance(pads, RPCPads):
            pads = RPCPads(pads)
        self.pads = pads

        self.memtype     = memtype     = "RPC"
        self.nranks      = nranks      = 1
        self.databits    = databits    = 16
        self.addressbits = addressbits = 12
        self.bankbits    = bankbits    = 2
        self.nphases     = nphases     = 4
        self.tck         = tck         = 1 / (nphases*sys_clk_freq)

        # CSRs -------------------------------------------------------------------------------------
        bitslip_cycles = 1
        self._rst                 = CSRStorage()
        self._dly_sel             = CSRStorage(databits//8)
        self._rdly_dq_bitslip_rst = CSR()
        self._rdly_dq_bitslip     = CSR()

        self._reset_done = CSRStatus()
        self._init_done  = CSRStatus()
        self._reset_fsm  = CSR()

        self._burst_stop = CSRStorage(reset=1)

        # PHY settings -----------------------------------------------------------------------------
        def get_cl(tck):
            # FIXME: for testing it's easier to use CL=8; read/write will be on phase 3; max sys_clk_freq=100e6
            return 8
            # tck is for DDR frequency
            f_to_cl = OrderedDict()
            f_to_cl[533e6]  =  3
            f_to_cl[800e6]  =  8
            f_to_cl[1200e6] =  8
            f_to_cl[1333e6] = 10
            f_to_cl[1600e6] = 11
            f_to_cl[1866e6] = 13
            for f, cl in f_to_cl.items():
                if tck >= 2/f:
                    return cl
            raise ValueError(tck)

        # RPC always has AL=1 and both read and write latencies are equal: RL=WL=AL+CL
        al = 1
        cwl = cl = get_cl(tck) + al

        cl_sys_latency  = get_sys_latency(nphases, cl)
        cwl_sys_latency = get_sys_latency(nphases, cwl)

        rdphase = get_sys_phase(nphases, cl_sys_latency, cl)
        wrphase = get_sys_phase(nphases, cwl_sys_latency, cwl)

        # Read latency
        db_cmd_dly   = 2  # (need 1 cycle to insert STB preamble + 1 more to always meet tCSS)
        cmd_ser_dly  = write_ser_latency
        read_mux_dly = 1
        bitslip_dly  = bitslip_cycles
        # Time until first data is available on DB
        read_db_dly = db_cmd_dly + cmd_ser_dly + cl_sys_latency
        # Time until data is deserialized (data present on 1ck signal)
        read_db_des_dly = read_db_dly + read_des_latency
        # Time until data is set on DFI (+1 because all data is present only on 2nd cycle)
        read_dfi_dly = read_mux_dly + bitslip_dly + 1
        # Final latency
        read_latency = read_db_des_dly + read_dfi_dly

        # Write latency for the controller. We must send 1 cycles of data mask before the
        # data, and we serialize data over 2 sysclk cycles due to minimal BL=16, so we
        # are writing in the 2 cycles following the cycle when we obtain data on DFI.
        # Other PHYs can send everything in 1 sysclk. Because of this spcific delay, we
        # have to increate tWR in the RPC SDRModule definition to meet tWR requirements.
        # +1 cycle needed to insert CS before command
        write_latency = cwl_sys_latency + 1

        self.settings = PhySettings(
            phytype       = phytype,
            memtype       = memtype,
            databits      = databits,
            dfi_databits  = 4*databits,
            nranks        = nranks,
            nphases       = nphases,
            rdphase       = rdphase,
            wrphase       = wrphase,
            cl            = cl,
            cwl           = cwl,
            read_latency  = read_latency,
            write_latency = write_latency,
            bitslips      = bitslip_cycles*2*2*nphases,
        )

        # DFI Interface ----------------------------------------------------------------------------
        # minimal BL=16, which gives 16*16=256 bits; with 4 phases we need 16/4=4 data widths
        dfi_params = dict(addressbits=addressbits, bankbits=bankbits, nranks=nranks,
                          databits=4*databits, nphases=nphases)

        # Register DFI history (from newest to oldest), as we need to operate on 3 subsequent cycles
        # hist[0] = dfi[N], hist[1] = dfi[N-1], ...
        self.dfi = dfi = DFIInterface(**dfi_params)
        dfi_hist = [dfi, DFIInterface(**dfi_params), DFIInterface(**dfi_params)]
        self.sync += dfi_hist[0].connect(dfi_hist[1], omit={"rddata", "rddata_valid"})
        self.sync += dfi_hist[1].connect(dfi_hist[2], omit={"rddata", "rddata_valid"})

        # Serialization ----------------------------------------------------------------------------
        # We have the following signals that have to be serialized:
        # - CLK (O)  - full-rate clock
        # - CS  (O)  - chip select
        # - STB (O)  - serial commands, serial preamble
        # - DB  (IO) - transmits data in/out, data mask, parallel commands
        # - DQS (IO) - strobe for commands/data/mask on DB pins
        # DQS is edge-aligned to CLK, while DB and STB are center-aligned to CLK (phase = -90).
        # Sending a parallel command (on DB pins):
        #  CLK: ____----____----____----____----____----____
        #  STB: ----------________________------------------
        #  DQS: ....................----____----____........
        #  DB:  ..........................PPPPnnnn..........
        # The signals prepared by BasePHY will all be phase-aligned. The concrete PHY should shift
        # them so that DB/STB/CS are delayed by 90 degrees in relation to CLK/DQS.

        # Signal values (de)serialized during 1 sysclk.
        # These signals must be populated in specific PHY implementations.
        self.clk_1ck_out  = clk_1ck_out  = Signal(2*nphases)
        self.stb_1ck_out  = stb_1ck_out  = Signal(2*nphases)
        self.cs_n_1ck_out = cs_n_1ck_out = Signal(2*nphases)

        self.dqs_1ck_out = dqs_1ck_out = Signal(2*nphases)
        self.dqs_1ck_in  = dqs_1ck_in  = Signal(2*nphases)
        self.dqs_oe      = dqs_oe      = Signal()

        self.db_1ck_out  = db_1ck_out  = [Signal(2*nphases) for _ in range(databits)]
        self.db_1ck_in   = db_1ck_in   = [Signal(2*nphases) for _ in range(databits)]
        self.db_oe       = db_oe       = Signal()

        # Clocks -----------------------------------------------------------------------------------
        self.comb += clk_1ck_out.eq(bitpattern("-_-_-_-_"))

        # DB muxing --------------------------------------------------------------------------------
        # Commands allowed by FSM
        cmd_valid = Signal()

        # Muxed cmd/data/mask
        db_1ck_data = [Signal(2*nphases) for _ in range(databits)]
        db_1ck_mask = [Signal(2*nphases) for _ in range(databits)]
        db_1ck_cmd  = [Signal(2*nphases) for _ in range(databits)]
        dq_data_en  = Signal()
        dq_mask_en  = Signal()
        dq_cmd_en   = Signal()
        dq_read_stb = Signal()

        # Output enable when writing cmd/data/mask
        # Mask is being send during negative half of sysclk
        self.comb += db_oe.eq(cmd_valid & (dq_data_en | dq_mask_en | dq_cmd_en))

        # Mux between cmd/data/mask
        for i in range(databits):
            self.comb += \
                If(dq_data_en,
                    db_1ck_out[i].eq(db_1ck_data[i])
                ).Elif(dq_mask_en,
                    db_1ck_out[i].eq(db_1ck_mask[i])
                ).Else(
                    db_1ck_out[i].eq(db_1ck_cmd[i])
                )

        # Parallel commands ------------------------------------------------------------------------
        # We need to insert 2 full-clk cycles of STB=0 before any command, to mark the beginning of
        # Request Packet. For that reason we use the previous values of DFI commands. To always be
        # able to meet tCSS, we have to add a delay of 1 more sysclk.
        # list from oldest to newest: dfi[N-1][p0], dfi[N-1][p1], ..., dfi[N][p0], dfi[N][p1], ...
        dfi_adapters = []
        for phase in dfi_hist[2].phases + dfi_hist[1].phases + dfi_hist[0].phases:
            adapter = DFIAdapter(phase)
            self.submodules += adapter
            dfi_adapters.append(adapter)
            self.comb += [
                # We always send one WORD, which consists of 32 bytes.
                adapter.bc.eq(0),
                # Always use fast refresh (equivalent to auto refresh) instead of low-power refresh
                # (equivalent to self refresh).
                adapter.ref_op.eq(adapter.REF_OP["FST"]),
            ]

        # Serialize commands to DB pins
        for i in range(databits):
            # A list of differential DB values using previous DFI coomand:
            # db_p[p][i], db_n[p][i], db_p[p+1][i], db_n[p+1][i], ...
            bits = [db for a in dfi_adapters[:nphases] for db in [a.db_p[i], a.db_n[i]]]
            self.comb += db_1ck_cmd[i].eq(Cat(*bits))

        # Commands go on the 2nd cycle, so use previous DFI
        self.comb += dq_cmd_en.eq(reduce(or_, [a.cmd_valid for a in dfi_adapters[:nphases]]))

        # Power Up Reset ---------------------------------------------------------------------------
        # During Power Up, after stabilizing clocks, Power Up Reset must be done. It consists of a
        # a Parallel Reset followed by two Serial Resets (2x8=16 full-rate cycles = 4 sys cycles).
        # We use an FSM to make sure that we pass only the commands from the controller that are
        # supported in the current state.
        t_reset          = 5e-6
        t_zqcinit        = 1e-6
        serial_reset_len = 4

        stb_reset_seq      = Signal()
        serial_reset_count = Signal(max=serial_reset_len + 1)

        # prolong the cmd_valid for the cmd latency (length of history)
        self.submodules.cmd_valid_sr = ShiftRegister(len(dfi_adapters) // nphases)
        self.comb += cmd_valid.eq(reduce(or_, self.cmd_valid_sr))

        self.submodules.reset_timer = WaitTimer(ceil(t_reset * sys_clk_freq))
        self.submodules.zqcinit_timer = WaitTimer(ceil(t_zqcinit * sys_clk_freq))

        self.submodules.reset_fsm = fsm = FSM()
        fsm.act("IDLE",
            NextValue(serial_reset_count, 0),
            NextValue(self._reset_done.status, 0),
            self.cmd_valid_sr.i.eq(dfi_adapters[2*nphases+0].is_cmd("RESET")),
            If(self.cmd_valid_sr.i,
                NextState("RESET_RECEIVED")
            ),
        )
        fsm.act("RESET_RECEIVED",
            NextState("SERIAL_RESET")
        )
        fsm.act("SERIAL_RESET",
            self.reset_timer.wait.eq(1),
            If(serial_reset_count != serial_reset_len,
               stb_reset_seq.eq(1),
               NextValue(serial_reset_count, serial_reset_count + 1),
            ),
            If(self.reset_timer.done,
                NextValue(self._reset_done.status, 1),
                NextState("RESET_DONE")
            )
        )
        fsm.act("RESET_DONE",
            self.cmd_valid_sr.i.eq(dfi_adapters[2*nphases+0].is_cmd(["PRE", "MRS", "ZQC_INIT"])),
            If(dfi_adapters[2*nphases+0].is_cmd("ZQC_INIT"),
                NextState("ZQC_INIT")
            )
        )
        fsm.act("ZQC_INIT",
            self.zqcinit_timer.wait.eq(1),
            If(self.zqcinit_timer.done,
                NextState("READY")
            )
        )
        fsm.act("READY",
            self._init_done.status.eq(1),
            self.cmd_valid_sr.i.eq(1),
            If(dfi_adapters[2*nphases+0].is_cmd("UTR") & (dfi_adapters[2*nphases+0].utr_en == 1),
                NextState("UTR_MODE")
            ),
            If(self._reset_fsm.re,
                NextState("IDLE")
            )
        )
        fsm.act("UTR_MODE",
            self._init_done.status.eq(1),
            self.cmd_valid_sr.i.eq(reduce(or_, [dfi_adapters[p].is_cmd(["UTR", "RD"])
                                                for p in [2*nphases, 2*nphases+self.settings.rdphase]])),
            If(dfi_adapters[2*nphases+0].is_cmd("UTR") & (dfi_adapters[2*nphases+0].utr_en == 0),
                NextState("READY")
            )
        )

        # STB --------------------------------------------------------------------------------------
        # Currently not sending any serial commands, but the STB pin must be held low for 2 full
        # rate cycles before writing a parallel command to activate the DRAM.
        stb_bits = []

        assert self.settings.rdphase == 3
        read_sent = [Signal(), Signal()]
        self.comb += read_sent[0].eq(dfi_adapters[3].cmd_valid & (dfi_adapters[3].is_cmd("RD") | dfi_adapters[3].is_cmd("WR")))
        self.sync += read_sent[1].eq(read_sent[0])

        for p in range(nphases):
            # Use cmd from current and prev cycle, depending on which phase the command appears on
            preamble = (dfi_adapters[p+2].cmd_valid | dfi_adapters[p+1].cmd_valid) & cmd_valid

            # force 000100xxxxxxxxxx after preamble to generate Burst Stop
            assert self.settings.rdphase == 3
            burst_stop = {
                0: [0, 0],
                1: [0, 1],
                2: [0, 0],
                3: [1, 1],
            }[(p+1)%4]
            read = read_sent[0] if p == 3 else read_sent[1]
            burst_stop_zero = [cmd_valid & read & (bs == 0) & self._burst_stop.storage
                               for bs in burst_stop]

            # We only want to use STB to start parallel commands, serial reset or to send NOPs. NOP
            # is indicated by the first two bits being high (0b11, and other as "don't care"), so
            # we can simply hold STB high all the time and reset is zeros for 8 cycles (1 sysclk).
            # stb_bits += 2 * [~(preamble | stb_reset_seq)]
            stb_bits += [
                ~(preamble | stb_reset_seq | burst_stop_zero[0]),
                ~(preamble | stb_reset_seq | burst_stop_zero[1]),
            ]

        self.comb += stb_1ck_out.eq(Cat(*stb_bits))

        # Chip Select ------------------------------------------------------------------------------
        # RPC has quite high required time of CS# low before sending a command (tCSS), this means
        # that we would need 1 more cmd_latency to support it for all standard frequencies. To meet
        # tCSH we hold CS# low 1 cycle after each command (and for writes until the burst ends).
        tCSS = 10e-9
        tCSH =  5e-9
        # CS# is held for 2 sysclks before any command, and 1 sysclk after any command
        assert 2 * 1/sys_clk_freq >= tCSS, "tCSS not met for commands on phase 0"
        assert 1 * 1/sys_clk_freq >= tCSH, "tCSH not met for commands on phase 3"

        cs         = Signal()
        cs_burst_hold = Signal()
        self.submodules.cs_hold = ShiftRegister(2)

        # FIXME: currently we hold CS low constantly to avoid problems with signal integrity on our board
        # lock CS when DFI sends cs_n=0 only on phase 0 (avoids start condition problems)
        cs_lock = Signal()
        cs_lock_cond = reduce(and_, [dfi_hist[0].phases[p].cs_n for p in range(1, nphases)])
        cs_lock_cond = cs_lock_cond & ~dfi_hist[0].p0.cs_n
        self.sync += If(cs_lock_cond, cs_lock.eq(1)).Elif(self._reset_fsm.re, cs_lock.eq(0))

        _any_cmd_valid = reduce(or_, (a.cmd_valid for a in dfi_adapters))
        self.comb += [
            # self.cs_hold.i.eq(cmd_valid & (_any_cmd_valid | cs_burst_hold)),
            # cs.eq(reduce(or_, self.cs_hold.sr)),
            # cs_n_1ck_out.eq(Replicate(~cs, len(cs_n_1ck_out))),
            cs_n_1ck_out.eq(Replicate(~cs_lock, len(cs_n_1ck_out))),
        ]

        # Data IN ----------------------------------------------------------------------------------
        # Synchronize the deserializer because we deserialize over 2 cycles.
        dq_in_cnt = Signal()
        self.sync += If(dq_read_stb, dq_in_cnt.eq(~dq_in_cnt)).Else(dq_in_cnt.eq(0))

        # Deserialize read data
        # sys_clk:    ------------____________------------____________
        # sysx4_clk:  ---___---___---___---___---___---___---___---___
        # DB num:     <0><1><2><3><4><5><6><7><8><9><a><b><c><d><e><f>
        for i in range(databits):
            # BL=16 -> 2ck
            n_1ck = 2*nphases
            rbits_2ck = Signal(2*n_1ck)
            rbits_1ck = Signal(n_1ck)

            self.comb += rbits_1ck.eq(db_1ck_in[i])
            self.sync += Case(dq_in_cnt, {
                0: rbits_2ck[:n_1ck].eq(rbits_1ck),
                1: rbits_2ck[n_1ck:].eq(rbits_1ck),
            })

            bs = BitSlip(len(rbits_2ck), cycles=bitslip_cycles,
                rst = self.get_rst(i//8, self._rdly_dq_bitslip_rst.re),
                slp = self.get_inc(i//8, self._rdly_dq_bitslip.re),
            )
            self.submodules += bs
            self.comb += bs.i.eq(rbits_2ck)

            for p in range(nphases):
                self.comb += [
                    dfi.phases[p].rddata[i+0*databits].eq(bs.o[p*nphases+0]),
                    dfi.phases[p].rddata[i+1*databits].eq(bs.o[p*nphases+1]),
                    dfi.phases[p].rddata[i+2*databits].eq(bs.o[p*nphases+2]),
                    dfi.phases[p].rddata[i+3*databits].eq(bs.o[p*nphases+3]),
                ]

        # Data OUT ---------------------------------------------------------------------------------
        # TODO: add 1 to tWR bacause we need 2 cycles to send data from 1 cycle

        # Before sending the actual data we have to send 2 32-bit data masks (4 DDR cycles). In the
        # mask each 0 bit means "write byte" and 1 means "mask byte". The 1st 32-bits mask the first
        # data WORD (32 bytes), and the 2nd 32-bits mask the last data WORD. Because we always send
        # 1 WORD of data (BC=0), we don't care about the 2nd mask (can send 1).
        #
        # Write data
        # DFI valid:  xxxxxxxxxxxxxxxxxx
        # sys_clk:    ------____________------------____________------------____________
        # sysx4_clk:  ---___---___---___---___---___---___---___---___---___---___---___
        # DB num:           <M><M><M><M><0><1><2><3><4><5><6><7><8><9><a><b><c><d><e><f>

        # Synchronize to 2 cycles, reset counting when dq_data_en turns high.
        db_cnt = Signal()
        self.sync += If(dq_data_en, db_cnt.eq(~db_cnt)).Else(db_cnt.eq(0))

        for i in range(databits):
            # Write data ---------------------------------------------------------------------------
            wbits = []
            for p in range(nphases):
                _dfi = dfi_hist[1] if p < nphases//2 else dfi_hist[2]
                wbits += [
                    _dfi.phases[p].wrdata[i+0*databits],
                    _dfi.phases[p].wrdata[i+1*databits],
                    _dfi.phases[p].wrdata[i+2*databits],
                    _dfi.phases[p].wrdata[i+3*databits],
                ]

            # Mux datas from 2 cycles to always serialize single cycle.
            wbits_2ck = [Cat(*wbits[:2*nphases]), Cat(*wbits[2*nphases:])]
            wbits_1ck = Signal(2*nphases)
            self.comb += wbits_1ck.eq(Array(wbits_2ck)[db_cnt])
            self.comb += db_1ck_data[i].eq(Cat(*wbits_1ck))

            # Data mask ----------------------------------------------------------------------------
            mask = [
                Constant(0),                                    # phase 0
                Constant(0),                                    # phase 0
                Constant(0),                                    # phase 1
                Constant(0),                                    # phase 1
                dfi_hist[0].phases[i//8 + 0].wrdata_mask[i%8],  # WL-2
                dfi_hist[0].phases[i//8 + 2].wrdata_mask[i%8],  # WL-2
                Constant(1),                                    # WL-1
                Constant(1),                                    # WL-1
            ]
            self.comb += db_1ck_mask[i].eq(Cat(*mask))

        # DQS --------------------------------------------------------------------------------------
        # Strobe pattern can go over 2 sysclk (do not count while transmitting mask)
        dqs_cnt = Signal()
        self.sync += If(dqs_oe & (~dq_mask_en), dqs_cnt.eq(~dqs_cnt)).Else(dqs_cnt.eq(0))

        pattern_2ck = [Signal(2*nphases), Signal(2*nphases)]
        self.comb += dqs_1ck_out.eq(Array(pattern_2ck)[dqs_cnt])

        # To avoid having to serialize dqs_oe, we serialize predefined pattern on dqs_out
        # All the patterns will be shifted back 90 degrees!
        data_pattern = [bitpattern("-_-_-_-_"), bitpattern("-_-_-_-_")]
        mask_pattern = [bitpattern("__-_-_-_"), bitpattern("-_-_-_-_")]
        phase_patterns = {
            0: [bitpattern("______-_"), bitpattern("-_______")],
            1: [bitpattern("________"), bitpattern("-_-_____")],
            2: [bitpattern("________"), bitpattern("__-_-___")],
            3: [bitpattern("________"), bitpattern("____-_-_")],
        }

        pattern_cases = \
            If(dq_mask_en,
                pattern_2ck[0].eq(mask_pattern[0]),
                pattern_2ck[1].eq(mask_pattern[1]),
            ).Elif(dq_data_en,
                pattern_2ck[0].eq(data_pattern[0]),
                pattern_2ck[1].eq(data_pattern[1]),
            ).Else(
                pattern_2ck[0].eq(0),
                pattern_2ck[1].eq(0),
            )
        any_phase_valid = 0
        for p in range(nphases):
            phase_valid = dfi_adapters[p].cmd_valid | dfi_adapters[nphases+p].cmd_valid
            pattern_cases = \
                If(phase_valid,
                    pattern_2ck[0].eq(phase_patterns[p][0]),
                    pattern_2ck[1].eq(phase_patterns[p][1]),
                ).Else(pattern_cases)
            any_phase_valid = any_phase_valid | phase_valid

        self.comb += pattern_cases
        self.comb += dqs_oe.eq(cmd_valid & (any_phase_valid | dq_mask_en | dq_data_en))

        # Read Control Path ------------------------------------------------------------------------
        # Creates a shift register of read commands coming from the DFI interface. This shift
        # register is used to indicate to the DFI interface that the read data is valid.
        self.submodules.rddata_en = rddata_en = ShiftRegister(self.settings.read_latency)
        self.comb += rddata_en.i.eq(dfi.phases[self.settings.rdphase].rddata_en)
        # -1 because of syncronious assignment
        self.sync += [phase.rddata_valid.eq(rddata_en[-1]) for phase in dfi.phases]
        # Strobe high when data from DRAM is available, before we can send it to DFI.
        self.sync += dq_read_stb.eq(rddata_en[read_db_des_dly-1] | rddata_en[read_db_des_dly-1 + 1])

        # Write Control Path -----------------------------------------------------------------------
        # Creates a shift register of write commands coming from the DFI interface. This shift
        # register is used to control DQ/DQS tristates.
        self.submodules.wrdata_en = wrdata_en = ShiftRegister(self.settings.write_latency + 2 + 1)
        self.comb += wrdata_en.i.eq(dfi.phases[self.settings.wrphase].wrdata_en & cmd_valid)
        # DQS Preamble and data mask are transmitted 1 cycle before data, then 2 cycles of data
        self.comb += dq_mask_en.eq(wrdata_en[write_latency])
        self.comb += dq_data_en.eq(wrdata_en[write_latency + 1] | wrdata_en[write_latency + 2])
        # Hold CS# low until end of write burst (use a latch as there can only be 1 write at a time)

        cs_wr_hold = Signal()
        self.sync += If(wrdata_en[0], cs_wr_hold.eq(1)).Elif(wrdata_en[-1], cs_wr_hold.eq(0))
        cs_rd_hold = Signal()
        self.sync += If(rddata_en[0], cs_rd_hold.eq(1)).Elif(rddata_en[-1], cs_rd_hold.eq(0))
        self.comb += cs_burst_hold.eq(cs_wr_hold | cs_rd_hold)

        # Additional variables for LiteScope -------------------------------------------------------
        variables = ["dq_data_en", "dq_mask_en", "dq_cmd_en", "dq_read_stb", "dfi_adapters",
                     "dq_in_cnt", "db_cnt", "dqs_cnt", "rddata_en", "wrdata_en"]
        for v in variables:
            setattr(self, v, locals()[v])

    def get_rst(self, byte, rst):
        return (self._dly_sel.storage[byte] & rst) | self._rst.storage

    def get_inc(self, byte, inc):
        return self._dly_sel.storage[byte] & inc

    def do_finalize(self):
        self.do_clock_serialization(self.clk_1ck_out, self.pads.clk_p, self.pads.clk_n)
        self.do_stb_serialization(self.stb_1ck_out, self.pads.stb)
        self.do_dqs_serialization(self.dqs_1ck_out, self.dqs_1ck_in, self.dqs_oe,
                                  self.pads.dqs_p, self.pads.dqs_n)
        self.do_db_serialization(self.db_1ck_out, self.db_1ck_in, self.db_oe, self.pads.db)
        self.do_cs_serialization(self.cs_n_1ck_out, self.pads.cs_n)

    # I/O implementation ---------------------------------------------------------------------------

    def do_clock_serialization(self, clk_1ck_out, clk_p, clk_n):
        raise NotImplementedError("Serialize the full-rate clock with 90 deg phase delay")

    def do_stb_serialization(self, stb_1ck_out, stb):
        raise NotImplementedError("Serialize the STB line")

    def do_dqs_serialization(self, dqs_1ck_out, dqs_1ck_in, dqs_oe, dqs_p, dqs_n):
        raise NotImplementedError("Tristate and (de)serialize DQS with 90 deg phase delay")

    def do_db_serialization(self, db_1ck_out, db_1ck_in, db_oe, db):
        raise NotImplementedError("Tristate and (de)serialize DB")

    def do_cs_serialization(self, cs_n_1ck_out, cs_n):
        raise NotImplementedError("Serialize the chip select line (CS#)")



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2016-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

"""Wishbone frontend for LiteDRAM"""

from math import log2

from migen import *

from litex.soc.interconnect import stream
from litedram.common import LiteDRAMNativePort
from litedram.frontend.adapter import LiteDRAMNativePortConverter


# LiteDRAMWishbone2Native --------------------------------------------------------------------------

class LiteDRAMWishbone2Native(Module):
    def __init__(self, wishbone, port, base_address=0x00000000):
        wishbone_data_width = len(wishbone.dat_w)
        port_data_width     = 2**int(log2(len(port.wdata.data))) # Round to lowest power 2
        ratio               = wishbone_data_width/port_data_width

        if wishbone_data_width != port_data_width:
            if wishbone_data_width > port_data_width:
                addr_shift = -log2_int(wishbone_data_width//port_data_width)
            else:
                addr_shift = log2_int(port_data_width//wishbone_data_width)
            new_port = LiteDRAMNativePort(
                mode          = port.mode,
                address_width = port.address_width + addr_shift,
                data_width    = wishbone_data_width
            )
            self.submodules += LiteDRAMNativePortConverter(new_port, port)
            port = new_port

        # # #

        aborted = Signal()
        offset  = base_address >> log2_int(port.data_width//8)

        self.submodules.fsm = fsm = FSM(reset_state="CMD")
        self.comb += [
            port.cmd.addr.eq(wishbone.adr - offset),
            port.cmd.we.eq(wishbone.we),
            port.cmd.last.eq(~wishbone.we), # Always wait for reads.
            port.flush.eq(~wishbone.cyc)    # Flush writes when transaction ends.
        ]
        fsm.act("CMD",
            port.cmd.valid.eq(wishbone.cyc & wishbone.stb),
            If(port.cmd.valid & port.cmd.ready &  wishbone.we, NextState("WRITE")),
            If(port.cmd.valid & port.cmd.ready & ~wishbone.we, NextState("READ")),
            NextValue(aborted, 0),
        )
        self.comb += [
            port.wdata.valid.eq(wishbone.stb & wishbone.we),
            If(ratio <= 1, If(~fsm.ongoing("WRITE"), port.wdata.valid.eq(0))),
            port.wdata.data.eq(wishbone.dat_w),
            port.wdata.we.eq(wishbone.sel),
        ]
        fsm.act("WRITE",
            NextValue(aborted, ~wishbone.cyc | aborted),
            If(port.wdata.valid & port.wdata.ready,
                wishbone.ack.eq(wishbone.cyc & ~aborted),
                NextState("CMD")
            ),
        )
        self.comb += port.rdata.ready.eq(1)
        fsm.act("READ",
            NextValue(aborted, ~wishbone.cyc | aborted),
            If(port.rdata.valid,
                wishbone.ack.eq(wishbone.cyc & ~aborted),
                wishbone.dat_r.eq(port.rdata.data),
                NextState("CMD")
            )
        )



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *


class ModeRegister:
    """RPC Mode Register encoding (RPC has only 1 mode register)"""
    def __init__(self):
        self.cl      = Signal(3)
        # TODO: in LPDDR3 nWR is the number of clock cycles determining when to start internal
        # precharge for a write burst when auto-precharge is enabled (ceil(tRW/tCK) ?)
        self.nwr     = Signal(3)
        self.zout    = Signal(4)
        self.odt     = Signal(3)
        self.odt_stb = Signal(1)
        self.csr_fx  = Signal(1)
        self.odt_pd  = Signal(1)
        self.tm      = Signal(1)

    CL = {
        8:  0b000,  # default
        10: 0b001,
        11: 0b010,
        13: 0b011,
        3:  0b110,
    }
    NWR = {
        4:  0b000,
        6:  0b001,
        7:  0b010,
        8:  0b011,  # default
        10: 0b100,
        12: 0b101,
        14: 0b110,
        16: 0b111,
    }
    ZOUT = {  # resistance in Ohms
        120:     0b0010,
        90:      0b0100,
        51.4:    0b0110,
        60:      0b1000,
        40:      0b1010,
        36:      0b1100,
        27.7:    0b1110,
        "short": 0b0001,  # 0bxxx1
        "open":  0b0000,  # output disabled, default
    }
    ODT = {
        60:     0b001,
        45:     0b010,
        25.7:   0b011,
        30:     0b100,
        20:     0b101,
        18:     0b110,
        13.85:  0b111,
        "open": 0b000,
    }

    # Encode mode register information in DFI address/bank
    DFI_ENCODING = {
        # field: (dfi_signal, width, offset)
        "cl":      ("address", 3,  0),
        # FIXME: not enough bits in DFI to store all data
        "nwr":     None,  # ("address", 3,  3),
        "zout":    ("address", 4,  3),
        "odt":     ("address", 3,  7),
        "csr_fx":  ("address", 1, 10),
        "odt_stb": ("bank",    1,  0),
        "odt_pd":  ("bank",    1,  1),
        "tm":      None,
    }

    @classmethod
    def dfi_encode(cls, **kwargs):
        address = 0
        bank = 0
        for field, encoding in cls.DFI_ENCODING.items():
            if encoding is None:
                continue
            sig, width, offset = encoding
            value = (kwargs[field] & (2**width - 1)) << offset
            if sig == "address":
                address |= value
            elif sig == "bank":
                bank |= value
            else:
                raise ValueError(sig)
        return address, bank

    def dfi_decode(self, dfi_phase):
        r = []
        for field, encoding in self.DFI_ENCODING.items():
            if encoding is None:
                continue
            sig, width, offset = encoding
            r += [getattr(self, field).eq(getattr(dfi_phase, sig)[offset:offset+width])]
        return r


class DFIAdapter(Module):
    """Translate DFI commands into RPC parallel commands format (Request Packet)

    Maps DFI commands to RPC parallel packet commands (sent over DB lines). Some commands cannot
    be represented by cas_n/ras_n/we_n combinations, for that reason when reset_n=0, some DFI
    commands are interpreted specially:
    - ACT -> perform a RESET sequence
    - MRS -> write UTR, utr_op and utr_en are encoded in phase.address (see `dfi_utr_encode`)
    - ZQC -> will perform di_ferent types of ZQ calibration: "init" when auto_precharge=1,
             "reset" if auto_precharge=0
    """
    ZQC_OP = {
        "init":  0b00,  # calibration after initialization
        "long":  0b01,
        "short": 0b10,
        "reset": 0b11,  # ZQ reset
    }
    REF_OP = {
        "FST": 0b00,  # FST refresh: tREFi = 100ns
        "LP":  0b00,  # LP refresh:  tREFi = 3.2us
    }
    UTR_OP = {  # Utility Register read pattern
        "0101": 0b00,
        "1100": 0b01,
        "0011": 0b10,
        "1010": 0b11,
    }

    SPECIAL_CMDS = {
        "RESET":    "ACT",
        "UTR":      "MRS",
        "ZQC_INIT": "ZQC",
    }

    @classmethod
    def dfi_utr_encode(cls, utr_op, utr_en):
        if isinstance(utr_op, str):
            utr_op = cls.UTR_OP[utr_op]
        address = 0
        address |= (utr_en &  0b1) << 0
        address |= (utr_op & 0b11) << 1
        return address

    def __init__(self, phase):
        self.db_p = Signal(16)  # on positive edge
        self.db_n = Signal(16)  # on negative edge

        self.cmd_valid = Signal()
        self.do_reset  = Signal()  # force sending RESET command
        self.bc        = Signal(6)  # burst count (bs+1 = number of 32-byte words in the transfer)
        self.ref_op    = Signal(2)

        # # #

        self.phase_cmd = phase_cmd = Signal(3)
        self.comb += phase_cmd.eq(Cat(phase.cas_n, phase.ras_n, phase.we_n))

        def _cmd(cas, ras, we):
            assert cas in [0, 1] and ras in [0, 1] and we in [0, 1]
            return ((1 - cas) << 0) | ((1 - ras) << 1) | ((1 - we) << 2)

        self.dfi_cmds = dfi_cmds = {
            "NOP": _cmd(cas=0, ras=0, we=0),
            "ACT": _cmd(cas=0, ras=1, we=0),
            "RD":  _cmd(cas=1, ras=0, we=0),
            "WR":  _cmd(cas=1, ras=0, we=1),
            "PRE": _cmd(cas=0, ras=1, we=1),
            "REF": _cmd(cas=1, ras=1, we=0),
            "ZQC": _cmd(cas=0, ras=0, we=1),
            "MRS": _cmd(cas=1, ras=1, we=1),
        }

        # precharge/refresh use a bank bitmask, so PRECHARGE ALL uses 0b1111
        self.special_cmds = special_cmds = Signal()
        self.utr_en       = utr_en       = phase.address[0]
        utr_op         = phase.address[1:3]
        bk             = Signal(4)
        zqc_op         = Signal(2)
        auto_precharge = Signal()
        mr             = ModeRegister()

        self.comb += mr.dfi_decode(phase)
        self.comb += [
            special_cmds.eq(~phase.reset_n),
            auto_precharge.eq(phase.address[10]),
            If(auto_precharge,
                bk.eq(0b1111),
            ).Else(  # binary to one-hot encoded
                Case(phase.bank[:2], {i: bk.eq(1 << i) for i in range(4)}),
            ),
            If(special_cmds,
                If(auto_precharge,
                    zqc_op.eq(self.ZQC_OP["init"]),
                ).Else(
                    zqc_op.eq(self.ZQC_OP["reset"]),
                )
            ).Else(
                If(auto_precharge,
                    zqc_op.eq(self.ZQC_OP["long"]),
                ).Else(
                    zqc_op.eq(self.ZQC_OP["short"]),
                )
            ),
        ]

        cases = {
            "NOP": [
                self.db_p.eq(0),
                self.db_n.eq(0),
            ],
            "ACT": [
                self.db_p[0:2  +1].eq(0b101),
                self.db_p[3:4  +1].eq(phase.bank[:2]),
                self.db_n[0      ].eq(0),
                self.db_n[1:12 +1].eq(phase.address[:12]),  # row address
            ],
            "RD": [
                self.db_p[0:2   +1].eq(0b000),
                self.db_p[3:4   +1].eq(phase.bank[:2]),
                self.db_p[5:10  +1].eq(self.bc),
                self.db_p[13:15 +1].eq(phase.address[4:6 +1]),
                self.db_n[0       ].eq(0),
                self.db_n[13:15 +1].eq(phase.address[7:9 +1]),
            ],
            "WR": [
                self.db_p[0:2   +1].eq(0b001),
                self.db_p[3:4   +1].eq(phase.bank[:2]),
                self.db_p[5:10  +1].eq(self.bc),
                self.db_p[13:15 +1].eq(phase.address[4:6 +1]),
                self.db_n[0       ].eq(0),
                self.db_n[13:15 +1].eq(phase.address[7:9 +1]),
            ],
            "PRE": [
                self.db_p[6:9+1].eq(bk),
                self.db_p[0:2+1].eq(0b100),
                self.db_n[0    ].eq(0),
            ],
            "REF": [
                self.db_p[6:9+1].eq(bk),
                self.db_p[0:2+1].eq(0b110),
                self.db_n[1:2+3].eq(self.ref_op),
                self.db_n[0    ].eq(0),
            ],
            "ZQC": [
                self.db_p[0:2  +1].eq(0b001),
                self.db_p[14:15+1].eq(zqc_op),
                self.db_n[0      ].eq(1),
            ],
            "MRS": [
                self.db_p[0:2  +1].eq(0b010),
                self.db_p[3:15 +1].eq(Cat(mr.cl, mr.nwr, mr.zout, mr.odt)),
                self.db_n[0      ].eq(0),
                self.db_n[12:15+1].eq(Cat(mr.odt_stb, mr.csr_fx, mr.odt_pd, mr.tm)),
            ],
            "UTR": [
                self.db_p[0:2  +1].eq(0b111),
                self.db_p[3      ].eq(utr_en),
                self.db_p[4:5  +1].eq(utr_op),
                self.db_n[0      ].eq(0),
            ],
            "RESET": [
                self.db_p.eq(0),
                self.db_n.eq(1),
            ],
        }

        self.comb += [
            If(self._is_cmd("RESET"),
                cases["RESET"],
                self.cmd_valid.eq(1),
            ).Elif(self._is_cmd("UTR"),
                cases["UTR"],
                self.cmd_valid.eq(1),
            ).Else(
                self.cmd_valid.eq(~self._is_cmd("NOP")),
                Case(phase_cmd, {dfi_cmds[cmd]: cases[cmd] for cmd in dfi_cmds.keys()}),
            ),
        ]

    def _is_cmd(self, cmd):
        if isinstance(cmd, list):
            return reduce(or_, (self._is_cmd(c) for c in cmd))
        if cmd in ["RESET", "UTR", "ZQC_INIT"]:
            return self.special_cmds & (self.phase_cmd == self.dfi_cmds[self.SPECIAL_CMDS[cmd]])
        return self.phase_cmd == self.dfi_cmds[cmd]

    def is_cmd(self, cmd):
        return self._is_cmd(cmd) & self.cmd_valid



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.cdc import PulseSynchronizer

from litex.soc.interconnect.csr import *

from litedram.common import *
from litedram.phy.dfi import *

from litedram.phy.utils import delayed, Latency
from litedram.phy.lpddr4.basephy import DoubleRateLPDDR4PHY

from litedram.phy.s7common import S7Common

class S7LPDDR4PHY(DoubleRateLPDDR4PHY, S7Common):
    def __init__(self, pads, *, iodelay_clk_freq, with_odelay, **kwargs):
        self.iodelay_clk_freq = iodelay_clk_freq

        # DoubleRateLPDDR4PHY outputs half-width signals (comparing to LPDDR4PHY) in sys2x domain.
        # This allows us to use 8:1 DDR OSERDESE2/ISERDESE2 to (de-)serialize the data.
        super().__init__(pads,
            ser_latency = Latency(sys2x=1),  # OSERDESE2 8:1 DDR (4 full-rate clocks)
            des_latency = Latency(sys2x=2),  # ISERDESE2 NETWORKING
            phytype     = self.__class__.__name__,
            serdes_reset_cnt=-1,
            **kwargs
        )

        self.settings.delays = 32
        self.settings.write_leveling = True
        self.settings.write_latency_calibration = True
        self.settings.write_dq_dqs_training = True
        self.settings.read_leveling = True

        # Parameters -------------------------------------------------------------------------------
        # Calculate value of taps needed to shift a signal by 90 degrees.
        # Using iodelay_clk_freq of 300MHz/400MHz is only valid for -3 and -2/2E speed grades.
        # Note: this should be named sys16x, but using sys8x due to a name hard-coded in BIOS
        assert iodelay_clk_freq in [200e6, 300e6, 400e6]
        iodelay_tap_average = 1 / (2*32 * iodelay_clk_freq)
        half_sys8x_taps = math.floor(self.tck / (4 * iodelay_tap_average))
        assert half_sys8x_taps < 32, "Exceeded ODELAYE2 max value: {} >= 32".format(half_sys8x_taps)

        # Registers --------------------------------------------------------------------------------
        self._half_sys8x_taps = CSRStorage(5, reset=half_sys8x_taps)

        # delay control
        self._rdly_dq_rst  = CSR()
        self._rdly_dq_inc  = CSR()
        self._rdly_dqs_rst = CSR()
        self._rdly_dqs_inc = CSR()
        if with_odelay:
            self._cdly_rst     = CSR()
            self._cdly_inc     = CSR()
            self._wdly_dq_rst  = CSR()
            self._wdly_dq_inc  = CSR()
            self._wdly_dqs_rst = CSR()
            self._wdly_dqs_inc = CSR()

        def cdc(i):
            o = Signal()
            psync = PulseSynchronizer("sys", "sys2x")
            self.submodules += psync
            self.comb += [
                psync.i.eq(i),
                o.eq(psync.o),
            ]
            return o

        rdly_dq_rst  = cdc(self._rdly_dq_rst.re)
        rdly_dq_inc  = cdc(self._rdly_dq_inc.re)
        rdly_dqs_rst = cdc(self._rdly_dqs_rst.re)
        rdly_dqs_inc = cdc(self._rdly_dqs_inc.re)
        if with_odelay:
            cdly_rst     = cdc(self._cdly_rst.re) | self._rst.storage
            cdly_inc     = cdc(self._cdly_inc.re)
            wdly_dq_rst  = cdc(self._wdly_dq_rst.re)
            wdly_dq_inc  = cdc(self._wdly_dq_inc.re)
            wdly_dqs_rst = cdc(self._wdly_dqs_rst.re)
            wdly_dqs_inc = cdc(self._wdly_dqs_inc.re)

        # In theory we should only need to delay by 2 cycles, but sometimes it happened that
        # DQ/DMI were transmitted incomplete due to OE being asserted too late/released too
        # early. For this reason we add OE margin extending it 1 cycle before and 1 after.
        # For DQS we already have margin so there should be no need for this.
        def oe_delay_data(oe):
            oe_d = Signal()
            delay = TappedDelayLine(oe, 3)
            self.submodules += ClockDomainsRenamer("sys2x")(delay)
            self.comb += oe_d.eq(reduce(or_, delay.taps))
            return oe_d

        def oe_delay_dqs(oe):
            delay = TappedDelayLine(oe, 2)
            self.submodules += ClockDomainsRenamer("sys2x")(delay)
            return delay.output

        # Serialization ----------------------------------------------------------------------------

        # Clock
        clk_dly = Signal()
        clk_ser = Signal()
        # Invert clk to have it phase shifted in relation to CS/CA, because we serialize it with DDR,
        # rising edge will then be in the middle of a data bit.
        self.oserdese2_ddr(din=~self.out.clk, dout=clk_ser if with_odelay else clk_dly, clk="sys8x")
        if with_odelay:
            self.odelaye2(din=clk_ser, dout=clk_dly, rst=cdly_rst, inc=cdly_inc)
        self.obufds(din=clk_dly, dout=self.pads.clk_p, dout_b=self.pads.clk_n)

        for cmd in ["cke", "odt", "reset_n"]:
            cmd_i = getattr(self.out, cmd)
            cmd_o = getattr(self.pads, cmd)
            cmd_ser = Signal()
            self.oserdese2_sdr(din=cmd_i, dout=cmd_ser if with_odelay else cmd_o, clk="sys8x")
            if with_odelay:
                self.odelaye2(din=cmd_ser, dout=cmd_o, rst=cdly_rst, inc=cdly_inc)

        # Commands
        cs_ser = Signal()
        if with_odelay:
            self.oserdese2_sdr(din=self.out.cs, dout=cs_ser, clk="sys8x")
            self.odelaye2(din=cs_ser, dout=self.pads.cs, rst=cdly_rst, inc=cdly_inc)
        else:
            self.oserdese2_sdr(din=self.out.cs, dout=self.pads.cs, clk="sys8x")
        for bit in range(6):
            ca_ser = Signal()
            if with_odelay:
                self.oserdese2_sdr(din=self.out.ca[bit], dout=ca_ser, clk="sys8x")
                self.odelaye2(din=ca_ser, dout=self.pads.ca[bit], rst=cdly_rst, inc=cdly_inc)
            else:
                self.oserdese2_sdr(din=self.out.ca[bit], dout=self.pads.ca[bit], clk="sys8x")

        # DQS
        for byte in range(self.databits//8):
            # DQS
            dqs_t     = Signal()
            dqs_ser   = Signal()
            dqs_dly   = Signal()
            dqs_i     = Signal()
            dqs_i_dly = Signal()
            # need to delay DQS if clocks are not phase aligned
            dqs_din = self.out.dqs_o[byte]
            if not with_odelay:
                dqs_din_d = Signal.like(dqs_din)
                self.sync.sys2x += dqs_din_d.eq(dqs_din)
                dqs_din = dqs_din_d
            self.oserdese2_ddr(
                din     = dqs_din,
                **(dict(dout_fb=dqs_ser) if with_odelay else dict(dout=dqs_dly)),
                tin     = ~oe_delay_dqs(self.out.dqs_oe),
                tout    = dqs_t,
                clk     = "sys8x" if with_odelay else "sys8x_90",
            )
            if with_odelay:
                self.odelaye2(
                    din  = dqs_ser,
                    dout = dqs_dly,
                    rst  = self.get_rst(byte, wdly_dqs_rst),
                    inc  = self.get_inc(byte, wdly_dqs_inc),
                    init = half_sys8x_taps,  # shifts by 90 degrees
                )
            self.iobufds(
                din      = dqs_dly,
                dout     = dqs_i,
                tin      = dqs_t,
                dinout   = self.pads.dqs_p[byte],
                dinout_b = self.pads.dqs_n[byte],
            )
            self.idelaye2(
                din  = dqs_i,
                dout = dqs_i_dly,
                rst  = self.get_rst(byte, rdly_dqs_rst),
                inc  = self.get_inc(byte, rdly_dqs_inc),
            )
            self.iserdese2_ddr(
                din  = dqs_i_dly,
                dout = self.out.dqs_i[byte],
                clk  = "sys8x",
            )

        # DMI
        for byte in range(self.databits//8):
            dmi_t   = Signal()
            dmi_ser = Signal()
            dmi_dly = Signal()
            self.oserdese2_ddr(
                din     = self.out.dmi_o[byte],
                **(dict(dout_fb=dmi_ser) if with_odelay else dict(dout=dmi_dly)),
                tin     = ~oe_delay_data(self.out.dmi_oe),
                tout    = dmi_t,
                clk     = "sys8x",
            )
            if with_odelay:
                self.odelaye2(
                    din  = dmi_ser,
                    dout = dmi_dly,
                    rst  = self.get_rst(byte, wdly_dq_rst),
                    inc  = self.get_inc(byte, wdly_dq_inc),
                )
            self.iobuf(
                din    = dmi_dly,
                dout   = Signal(),
                tin    = dmi_t,
                dinout = self.pads.dmi[byte],
            )

        # DQ
        for bit in range(self.databits):
            dq_t     = Signal()
            dq_ser   = Signal()
            dq_dly   = Signal()
            dq_i     = Signal()
            dq_i_dly = Signal()
            self.oserdese2_ddr(
                din     = self.out.dq_o[bit],
                **(dict(dout_fb=dq_ser) if with_odelay else dict(dout=dq_dly)),
                tin     = ~oe_delay_data(self.out.dmi_oe),
                tout    = dq_t,
                clk     = "sys8x",
            )
            if with_odelay:
                self.odelaye2(
                    din  = dq_ser,
                    dout = dq_dly,
                    rst  = self.get_rst(bit//8, wdly_dq_rst),
                    inc  = self.get_inc(bit//8, wdly_dq_inc),
                )
            self.iobuf(
                din    = dq_dly,
                dout   = dq_i,
                dinout = self.pads.dq[bit],
                tin    = dq_t
            )
            self.idelaye2(
                din  = dq_i,
                dout = dq_i_dly,
                rst  = self.get_rst(bit//8, rdly_dq_rst),
                inc  = self.get_inc(bit//8, rdly_dq_inc)
            )
            self.iserdese2_ddr(
                din  = dq_i_dly,
                dout = self.out.dq_i[bit],
                clk  = "sys8x"
            )

# PHY variants -------------------------------------------------------------------------------------

class V7LPDDR4PHY(S7LPDDR4PHY):
    """Xilinx Virtex7 LPDDR4 PHY (with odelay)"""
    def __init__(self, pads, **kwargs):
        S7LPDDR4PHY.__init__(self, pads, with_odelay=True, **kwargs)

class K7LPDDR4PHY(S7LPDDR4PHY):
    """Xilinx Kintex7 LPDDR4 PHY (with odelay)"""
    def __init__(self, pads, **kwargs):
        S7LPDDR4PHY.__init__(self, pads, with_odelay=True, **kwargs)

class A7LPDDR4PHY(S7LPDDR4PHY):
    """Xilinx Artix7 LPDDR4 PHY (without odelay)

    This variant requires generating sys8x_90 clock in CRG with a 90° phase shift vs sys8x.
    """
    def __init__(self, pads, **kwargs):
        S7LPDDR4PHY.__init__(self, pads, with_odelay=False, **kwargs)
